# encoding: utf-8

from __future__ import unicode_literals, absolute_import

import os
import sys
import locale
from itertools import chain
from six import iterkeys, iteritems
from six.moves.configparser import ConfigParser

from .autocomplete import SIMPLE as default_completion, ALL_MODES


class Struct(object):
    """Simple class for instantiating objects we can add arbitrary attributes
    to and use for various arbitrary things."""


def getpreferredencoding():
    """Get the user's preferred encoding."""
    return locale.getpreferredencoding() or sys.getdefaultencoding()


def can_encode(c):
    try:
        c.encode(getpreferredencoding())
        return True
    except UnicodeEncodeError:
        return False


def supports_box_chars():
    """Check if the encoding supports Unicode box characters."""
    return all(map(can_encode, "│─└┘┌┐"))


def get_config_home():
    """Returns the base directory for bpython's configuration files."""
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME", "~/.config")
    return os.path.join(xdg_config_home, "bpython")


def default_config_path():
    """Returns bpython's default configuration file path."""
    return os.path.join(get_config_home(), "config")


def fill_config_with_default_values(config, default_values):
    for section in iterkeys(default_values):
        if not config.has_section(section):
            config.add_section(section)

        for (opt, val) in iteritems(default_values[section]):
            if not config.has_option(section, opt):
                config.set(section, opt, "%s" % (val,))


def loadini(struct, configfile):
    """Loads .ini configuration file and stores its values in struct"""

    config_path = os.path.expanduser(configfile)

    config = ConfigParser()
    defaults = {
        "general": {
            "arg_spec": True,
            "auto_display_list": True,
            "autocomplete_mode": default_completion,
            "color_scheme": "default",
            "complete_magic_methods": True,
            "dedent_after": 1,
            "default_autoreload": False,
            "editor": os.environ.get("VISUAL", os.environ.get("EDITOR", "vi")),
            "flush_output": True,
            "highlight_show_source": True,
            "hist_duplicates": True,
            "hist_file": "~/.pythonhist",
            "hist_length": 1000,
            "paste_time": 0.02,
            "pastebin_confirm": True,
            "pastebin_expiry": "1week",
            "pastebin_helper": "",
            "pastebin_url": "https://bpaste.net",
            "save_append_py": False,
            "single_undo_time": 1.0,
            "syntax": True,
            "tab_length": 4,
            "unicode_box": True,
        },
        "keyboard": {
            "backspace": "C-h",
            "beginning_of_line": "C-a",
            "clear_line": "C-u",
            "clear_screen": "C-l",
            "clear_word": "C-w",
            "copy_clipboard": "F10",
            "cut_to_buffer": "C-k",
            "delete": "C-d",
            "down_one_line": "C-n",
            "edit_config": "F3",
            "edit_current_block": "C-x",
            "end_of_line": "C-e",
            "exit": "",
            "external_editor": "F7",
            "help": "F1",
            "incremental_search": "M-s",
            "last_output": "F9",
            "left": "C-b",
            "pastebin": "F8",
            "redo": "C-g",
            "reimport": "F6",
            "reverse_incremental_search": "M-r",
            "right": "C-f",
            "save": "C-s",
            "search": "C-o",
            "show_source": "F2",
            "suspend": "C-z",
            "toggle_file_watch": "F5",
            "transpose_chars": "C-t",
            "undo": "C-r",
            "up_one_line": "C-p",
            "yank_from_buffer": "C-y",
        },
        "cli": {"suggestion_width": 0.8, "trim_prompts": False,},
        "curtsies": {"list_above": False, "right_arrow_completion": True,},
    }

    default_keys_to_commands = dict(
        (value, key) for (key, value) in iteritems(defaults["keyboard"])
    )

    fill_config_with_default_values(config, defaults)
    try:
        if not config.read(config_path):
            # No config file. If the user has it in the old place then complain
            if os.path.isfile(os.path.expanduser("~/.bpython.ini")):
                sys.stderr.write(
                    "Error: It seems that you have a config file at "
                    "~/.bpython.ini. Please move your config file to "
                    "%s\n" % default_config_path()
                )
                sys.exit(1)
    except UnicodeDecodeError as e:
        sys.stderr.write(
            "Error: Unable to parse config file at '{}' due to an "
            "encoding issue. Please make sure to fix the encoding "
            "of the file or remove it and then try again.\n".format(config_path)
        )
        sys.exit(1)

    def get_key_no_doublebind(command):
        default_commands_to_keys = defaults["keyboard"]
        requested_key = config.get("keyboard", command)

        try:
            default_command = default_keys_to_commands[requested_key]

            if default_commands_to_keys[default_command] == config.get(
                "keyboard", default_command
            ):
                setattr(struct, "%s_key" % default_command, "")
        except KeyError:
            pass

        return requested_key

    struct.config_path = config_path

    struct.dedent_after = config.getint("general", "dedent_after")
    struct.tab_length = config.getint("general", "tab_length")
    struct.auto_display_list = config.getboolean("general", "auto_display_list")
    struct.syntax = config.getboolean("general", "syntax")
    struct.arg_spec = config.getboolean("general", "arg_spec")
    struct.paste_time = config.getfloat("general", "paste_time")
    struct.single_undo_time = config.getfloat("general", "single_undo_time")
    struct.highlight_show_source = config.getboolean(
        "general", "highlight_show_source"
    )
    struct.hist_file = config.get("general", "hist_file")
    struct.editor = config.get("general", "editor")
    struct.hist_length = config.getint("general", "hist_length")
    struct.hist_duplicates = config.getboolean("general", "hist_duplicates")
    struct.flush_output = config.getboolean("general", "flush_output")
    struct.default_autoreload = config.getboolean(
        "general", "default_autoreload"
    )

    struct.pastebin_key = get_key_no_doublebind("pastebin")
    struct.copy_clipboard_key = get_key_no_doublebind("copy_clipboard")
    struct.save_key = get_key_no_doublebind("save")
    struct.search_key = get_key_no_doublebind("search")
    struct.show_source_key = get_key_no_doublebind("show_source")
    struct.suspend_key = get_key_no_doublebind("suspend")
    struct.toggle_file_watch_key = get_key_no_doublebind("toggle_file_watch")
    struct.undo_key = get_key_no_doublebind("undo")
    struct.redo_key = get_key_no_doublebind("redo")
    struct.reimport_key = get_key_no_doublebind("reimport")
    struct.reverse_incremental_search_key = get_key_no_doublebind(
        "reverse_incremental_search"
    )
    struct.incremental_search_key = get_key_no_doublebind("incremental_search")
    struct.up_one_line_key = get_key_no_doublebind("up_one_line")
    struct.down_one_line_key = get_key_no_doublebind("down_one_line")
    struct.cut_to_buffer_key = get_key_no_doublebind("cut_to_buffer")
    struct.yank_from_buffer_key = get_key_no_doublebind("yank_from_buffer")
    struct.clear_word_key = get_key_no_doublebind("clear_word")
    struct.backspace_key = get_key_no_doublebind("backspace")
    struct.clear_line_key = get_key_no_doublebind("clear_line")
    struct.clear_screen_key = get_key_no_doublebind("clear_screen")
    struct.delete_key = get_key_no_doublebind("delete")

    struct.left_key = get_key_no_doublebind("left")
    struct.right_key = get_key_no_doublebind("right")
    struct.end_of_line_key = get_key_no_doublebind("end_of_line")
    struct.beginning_of_line_key = get_key_no_doublebind("beginning_of_line")
    struct.transpose_chars_key = get_key_no_doublebind("transpose_chars")
    struct.exit_key = get_key_no_doublebind("exit")
    struct.last_output_key = get_key_no_doublebind("last_output")
    struct.edit_config_key = get_key_no_doublebind("edit_config")
    struct.edit_current_block_key = get_key_no_doublebind("edit_current_block")
    struct.external_editor_key = get_key_no_doublebind("external_editor")
    struct.help_key = get_key_no_doublebind("help")

    struct.pastebin_confirm = config.getboolean("general", "pastebin_confirm")
    struct.pastebin_url = config.get("general", "pastebin_url")
    struct.pastebin_expiry = config.get("general", "pastebin_expiry")
    struct.pastebin_helper = config.get("general", "pastebin_helper")

    struct.cli_suggestion_width = config.getfloat("cli", "suggestion_width")
    struct.cli_trim_prompts = config.getboolean("cli", "trim_prompts")

    struct.complete_magic_methods = config.getboolean(
        "general", "complete_magic_methods"
    )
    struct.autocomplete_mode = config.get("general", "autocomplete_mode")
    struct.save_append_py = config.getboolean("general", "save_append_py")

    struct.curtsies_list_above = config.getboolean("curtsies", "list_above")
    struct.curtsies_right_arrow_completion = config.getboolean(
        "curtsies", "right_arrow_completion"
    )

    color_scheme_name = config.get("general", "color_scheme")

    default_colors = {
        "keyword": "y",
        "name": "c",
        "comment": "b",
        "string": "m",
        "error": "r",
        "number": "G",
        "operator": "Y",
        "punctuation": "y",
        "token": "C",
        "background": "d",
        "output": "w",
        "main": "c",
        "paren": "R",
        "prompt": "c",
        "prompt_more": "g",
        "right_arrow_suggestion": "K",
    }

    if color_scheme_name == "default":
        struct.color_scheme = default_colors
    else:
        struct.color_scheme = dict()

        theme_filename = color_scheme_name + ".theme"
        path = os.path.expanduser(
            os.path.join(get_config_home(), theme_filename)
        )
        try:
            load_theme(struct, path, struct.color_scheme, default_colors)
        except EnvironmentError:
            sys.stderr.write(
                "Could not load theme '%s'.\n" % (color_scheme_name,)
            )
            sys.exit(1)

    # expand path of history file
    struct.hist_file = os.path.expanduser(struct.hist_file)

    # verify completion mode
    if struct.autocomplete_mode not in ALL_MODES:
        struct.autocomplete_mode = default_completion

    # set box drawing characters
    if config.getboolean("general", "unicode_box") and supports_box_chars():
        struct.left_border = "│"
        struct.right_border = "│"
        struct.top_border = "─"
        struct.bottom_border = "─"
        struct.left_bottom_corner = "└"
        struct.right_bottom_corner = "┘"
        struct.left_top_corner = "┌"
        struct.right_top_corner = "┐"
    else:
        struct.left_border = "|"
        struct.right_border = "|"
        struct.top_border = "-"
        struct.bottom_border = "-"
        struct.left_bottom_corner = "+"
        struct.right_bottom_corner = "+"
        struct.left_top_corner = "+"
        struct.right_top_corner = "+"


def load_theme(struct, path, colors, default_colors):
    theme = ConfigParser()
    with open(path, "r") as f:
        theme.readfp(f)
    for k, v in chain(theme.items("syntax"), theme.items("interface")):
        if theme.has_option("syntax", k):
            colors[k] = theme.get("syntax", k)
        else:
            colors[k] = theme.get("interface", k)

    # Check against default theme to see if all values are defined
    for k, v in iteritems(default_colors):
        if k not in colors:
            colors[k] = v



# encoding: utf-8

import re

from bpython.test import unittest
from bpython.line import (
    current_word,
    current_dict_key,
    current_dict,
    current_string,
    current_object,
    current_object_attribute,
    current_from_import_from,
    current_from_import_import,
    current_import,
    current_method_definition_name,
    current_single_word,
    current_expression_attribute,
    current_dotted_attribute,
)


def cursor(s):
    """'ab|c' -> (2, 'abc')"""
    cursor_offset = s.index("|")
    line = s[:cursor_offset] + s[cursor_offset + 1 :]
    return cursor_offset, line


def decode(s):
    """'a<bd|c>d' -> ((3, 'abcd'), (1, 3, 'bdc'))"""

    if not s.count("|") == 1:
        raise ValueError("match helper needs | to occur once")
    if s.count("<") != s.count(">") or s.count("<") not in (0, 1):
        raise ValueError("match helper needs <, and > to occur just once")
    matches = list(re.finditer(r"[<>|]", s))
    assert len(matches) in [1, 3], [m.group() for m in matches]
    d = {}
    for i, m in enumerate(matches):
        d[m.group(0)] = m.start() - i
        s = s[: m.start() - i] + s[m.end() - i :]
    assert len(d) in [1, 3], "need all the parts just once! %r" % d

    if "<" in d:
        return (d["|"], s), (d["<"], d[">"], s[d["<"] : d[">"]])
    else:
        return (d["|"], s), None


def line_with_cursor(cursor_offset, line):
    return line[:cursor_offset] + "|" + line[cursor_offset:]


def encode(cursor_offset, line, result):
    """encode(3, 'abdcd', (1, 3, 'bdc')) -> a<bd|c>d'

    Written for prettier assert error messages
    """
    encoded_line = line_with_cursor(cursor_offset, line)
    if result is None:
        return encoded_line
    start, end, value = result
    assert line[start:end] == value
    if start < cursor_offset:
        encoded_line = encoded_line[:start] + "<" + encoded_line[start:]
    else:
        encoded_line = (
            encoded_line[: start + 1] + "<" + encoded_line[start + 1 :]
        )
    if end < cursor_offset:
        encoded_line = encoded_line[: end + 1] + ">" + encoded_line[end + 1 :]
    else:
        encoded_line = encoded_line[: end + 2] + ">" + encoded_line[end + 2 :]
    return encoded_line


class LineTestCase(unittest.TestCase):
    def assertAccess(self, s):
        r"""Asserts that self.func matches as described
        by s, which uses a little language to describe matches:

        abcd<efg>hijklmnopqrstuvwx|yz
           /|\ /|\               /|\
            |   |                 |
         the function should   the current cursor position
         match this "efg"      is between the x and y
        """
        (cursor_offset, line), match = decode(s)
        result = self.func(cursor_offset, line)

        self.assertEqual(
            result,
            match,
            "%s(%r) result\n%r (%r) doesn't match expected\n%r (%r)"
            % (
                self.func.__name__,
                line_with_cursor(cursor_offset, line),
                encode(cursor_offset, line, result),
                result,
                s,
                match,
            ),
        )


class TestHelpers(LineTestCase):
    def test_I(self):
        self.assertEqual(cursor("asd|fgh"), (3, "asdfgh"))

    def test_decode(self):
        self.assertEqual(decode("a<bd|c>d"), ((3, "abdcd"), (1, 4, "bdc")))
        self.assertEqual(decode("a|<bdc>d"), ((1, "abdcd"), (1, 4, "bdc")))
        self.assertEqual(decode("a<bdc>d|"), ((5, "abdcd"), (1, 4, "bdc")))

    def test_encode(self):
        self.assertEqual(encode(3, "abdcd", (1, 4, "bdc")), "a<bd|c>d")
        self.assertEqual(encode(1, "abdcd", (1, 4, "bdc")), "a|<bdc>d")
        self.assertEqual(encode(4, "abdcd", (1, 4, "bdc")), "a<bdc|>d")
        self.assertEqual(encode(5, "abdcd", (1, 4, "bdc")), "a<bdc>d|")

    def test_assert_access(self):
        def dumb_func(cursor_offset, line):
            return (0, 2, "ab")

        self.func = dumb_func
        self.assertAccess("<a|b>d")


class TestCurrentWord(LineTestCase):
    def setUp(self):
        self.func = current_word

    def test_simple(self):
        self.assertAccess("|")
        self.assertAccess("|asdf")
        self.assertAccess("<a|sdf>")
        self.assertAccess("<asdf|>")
        self.assertAccess("<asdfg|>")
        self.assertAccess("asdf + <asdfg|>")
        self.assertAccess("<asdfg|> + asdf")

    def test_inside(self):
        self.assertAccess("<asd|>")
        self.assertAccess("<asd|fg>")

    def test_dots(self):
        self.assertAccess("<Object.attr1|>")
        self.assertAccess("<Object.attr1.attr2|>")
        self.assertAccess("<Object.att|r1.attr2>")
        self.assertAccess("stuff[stuff] + {123: 456} + <Object.attr1.attr2|>")
        self.assertAccess("stuff[<asd|fg>]")
        self.assertAccess("stuff[asdf[<asd|fg>]")

    def test_non_dots(self):
        self.assertAccess("].asdf|")
        self.assertAccess(").asdf|")
        self.assertAccess("foo[0].asdf|")
        self.assertAccess("foo().asdf|")
        self.assertAccess("foo().|")
        self.assertAccess("foo().asdf.|")
        self.assertAccess("foo[0].asdf.|")

    def test_open_paren(self):
        self.assertAccess("<foo(|>")
        # documenting current behavior - TODO is this intended?


class TestCurrentDictKey(LineTestCase):
    def setUp(self):
        self.func = current_dict_key

    def test_simple(self):
        self.assertAccess("asdf|")
        self.assertAccess("asdf|")
        self.assertAccess("asdf[<>|")
        self.assertAccess("asdf[<>|]")
        self.assertAccess("object.dict[<abc|>")
        self.assertAccess("asdf|")
        self.assertAccess("asdf[<(>|]")
        self.assertAccess("asdf[<(1>|]")
        self.assertAccess("asdf[<(1,>|]")
        self.assertAccess("asdf[<(1, >|]")
        self.assertAccess("asdf[<(1, 2)>|]")
        # TODO self.assertAccess('d[d[<12|>')
        self.assertAccess("d[<'a>|")


class TestCurrentDict(LineTestCase):
    def setUp(self):
        self.func = current_dict

    def test_simple(self):
        self.assertAccess("asdf|")
        self.assertAccess("asdf|")
        self.assertAccess("<asdf>[|")
        self.assertAccess("<asdf>[|]")
        self.assertAccess("<object.dict>[abc|")
        self.assertAccess("asdf|")


class TestCurrentString(LineTestCase):
    def setUp(self):
        self.func = current_string

    def test_closed(self):
        self.assertAccess('"<as|df>"')
        self.assertAccess('"<asdf|>"')
        self.assertAccess('"<|asdf>"')
        self.assertAccess("'<asdf|>'")
        self.assertAccess("'<|asdf>'")
        self.assertAccess("'''<asdf|>'''")
        self.assertAccess('"""<asdf|>"""')
        self.assertAccess('asdf.afd("a") + "<asdf|>"')

    def test_open(self):
        self.assertAccess('"<as|df>')
        self.assertAccess('"<asdf|>')
        self.assertAccess('"<|asdf>')
        self.assertAccess("'<asdf|>")
        self.assertAccess("'<|asdf>")
        self.assertAccess("'''<asdf|>")
        self.assertAccess('"""<asdf|>')
        self.assertAccess('asdf.afd("a") + "<asdf|>')


class TestCurrentObject(LineTestCase):
    def setUp(self):
        self.func = current_object

    def test_simple(self):
        self.assertAccess("<Object>.attr1|")
        self.assertAccess("<Object>.|")
        self.assertAccess("Object|")
        self.assertAccess("Object|.")
        self.assertAccess("<Object>.|")
        self.assertAccess("<Object.attr1>.attr2|")
        self.assertAccess("<Object>.att|r1.attr2")
        self.assertAccess("stuff[stuff] + {123: 456} + <Object.attr1>.attr2|")
        self.assertAccess("stuff[asd|fg]")
        self.assertAccess("stuff[asdf[asd|fg]")


class TestCurrentAttribute(LineTestCase):
    def setUp(self):
        self.func = current_object_attribute

    def test_simple(self):
        self.assertAccess("Object.<attr1|>")
        self.assertAccess("Object.attr1.<attr2|>")
        self.assertAccess("Object.<att|r1>.attr2")
        self.assertAccess("stuff[stuff] + {123: 456} + Object.attr1.<attr2|>")
        self.assertAccess("stuff[asd|fg]")
        self.assertAccess("stuff[asdf[asd|fg]")
        self.assertAccess("Object.attr1.<|attr2>")
        self.assertAccess("Object.<attr1|>.attr2")


class TestCurrentFromImportFrom(LineTestCase):
    def setUp(self):
        self.func = current_from_import_from

    def test_simple(self):
        self.assertAccess("from <sys|> import path")
        self.assertAccess("from <sys> import path|")
        self.assertAccess("if True|: from sys import path")
        self.assertAccess("if True: |from sys import path")
        self.assertAccess("if True: from <sys> import p|ath")
        self.assertAccess("if True: from sys imp|ort path")
        self.assertAccess("if True: from sys import |path")
        self.assertAccess("if True: from sys import path.stu|ff")
        self.assertAccess("if True: from <sys.path> import sep|")
        self.assertAccess("from <os.p|>")


class TestCurrentFromImportImport(LineTestCase):
    def setUp(self):
        self.func = current_from_import_import

    def test_simple(self):
        self.assertAccess("from sys import <path|>")
        self.assertAccess("from sys import <p|ath>")
        self.assertAccess("from sys import |path")
        self.assertAccess("from sys| import path")
        self.assertAccess("from s|ys import path")
        self.assertAccess("from |sys import path")
        self.assertAccess("from xml.dom import <N|ode>")
        # because syntax error
        self.assertAccess("from xml.dom import Node.as|d")


class TestCurrentImport(LineTestCase):
    def setUp(self):
        self.func = current_import

    def test_simple(self):
        self.assertAccess("import <path|>")
        self.assertAccess("import <p|ath>")
        self.assertAccess("import |path")
        self.assertAccess("import path, <another|>")
        self.assertAccess("import path another|")
        self.assertAccess("if True: import <path|>")
        self.assertAccess("if True: import <xml.dom.minidom|>")
        self.assertAccess("if True: import <xml.do|m.minidom>")
        self.assertAccess("if True: import <xml.do|m.minidom> as something")


class TestMethodDefinitionName(LineTestCase):
    def setUp(self):
        self.func = current_method_definition_name

    def test_simple(self):
        self.assertAccess("def <foo|>")
        self.assertAccess("    def bar(x, y)|:")
        self.assertAccess("    def <bar|>(x, y)")


class TestSingleWord(LineTestCase):
    def setUp(self):
        self.func = current_single_word

    def test_simple(self):
        self.assertAccess("foo.bar|")
        self.assertAccess(".foo|")
        self.assertAccess(" <foo|>")


class TestCurrentExpressionAttribute(LineTestCase):
    def setUp(self):
        self.func = current_expression_attribute

    def test_simple(self):
        self.assertAccess("Object.<attr1|>.")
        self.assertAccess("Object.<|attr1>.")
        self.assertAccess("Object.(|)")
        self.assertAccess("Object.another.(|)")
        self.assertAccess("asdf asdf asdf.(abc|)")

    def test_without_dot(self):
        self.assertAccess("Object|")
        self.assertAccess("Object|.")
        self.assertAccess("|Object.")

    def test_with_whitespace(self):
        self.assertAccess("Object. <attr|>")
        self.assertAccess("Object .<attr|>")
        self.assertAccess("Object . <attr|>")
        self.assertAccess("Object .asdf attr|")
        self.assertAccess("Object .<asdf|> attr")
        self.assertAccess("Object. asdf attr|")
        self.assertAccess("Object. <asdf|> attr")
        self.assertAccess("Object . asdf attr|")
        self.assertAccess("Object . <asdf|> attr")

    def test_indexing(self):
        self.assertAccess("abc[def].<ghi|>")
        self.assertAccess("abc[def].<|ghi>")
        self.assertAccess("abc[def].<gh|i>")
        self.assertAccess("abc[def].gh |i")
        self.assertAccess("abc[def]|")

    def test_strings(self):
        self.assertAccess('"hey".<a|>')
        self.assertAccess('"hey"|')
        self.assertAccess('"hey"|.a')
        self.assertAccess('"hey".<a|b>')
        self.assertAccess('"hey".asdf d|')
        self.assertAccess('"hey".<|>')


class TestCurrentDottedAttribute(LineTestCase):
    def setUp(self):
        self.func = current_dotted_attribute

    def test_simple(self):
        self.assertAccess("<obj.attr>|")
        self.assertAccess("(<obj.attr>|")
        self.assertAccess("[<obj.attr>|")
        self.assertAccess("m.body[0].value|")
        self.assertAccess("m.body[0].attr.value|")


if __name__ == "__main__":
    unittest.main()



# encoding: utf-8

from __future__ import absolute_import

import collections
import io
import logging
import sys
from optparse import Option

import curtsies
import curtsies.window
import curtsies.input
import curtsies.events

from .curtsiesfrontend.repl import BaseRepl
from .curtsiesfrontend.coderunner import SystemExitFromCodeRunner
from .curtsiesfrontend.interpreter import Interp
from . import args as bpargs
from . import translations
from .translations import _
from .importcompletion import find_iterator
from .curtsiesfrontend import events as bpythonevents
from . import inspection
from .repl import extract_exit_value
from ._py3compat import py3

logger = logging.getLogger(__name__)


repl = None  # global for `from bpython.curtsies import repl`
# WARNING Will be a problem if more than one repl is ever instantiated this way


class FullCurtsiesRepl(BaseRepl):
    def __init__(self, config, locals_, banner, interp=None):
        self.input_generator = curtsies.input.Input(
            keynames="curtsies", sigint_event=True, paste_threshold=None
        )
        self.window = curtsies.window.CursorAwareWindow(
            sys.stdout,
            sys.stdin,
            keep_last_line=True,
            hide_cursor=False,
            extra_bytes_callback=self.input_generator.unget_bytes,
        )

        self._request_refresh = self.input_generator.event_trigger(
            bpythonevents.RefreshRequestEvent
        )
        self._schedule_refresh = self.input_generator.scheduled_event_trigger(
            bpythonevents.ScheduledRefreshRequestEvent
        )
        self._request_reload = self.input_generator.threadsafe_event_trigger(
            bpythonevents.ReloadEvent
        )
        self.interrupting_refresh = self.input_generator.threadsafe_event_trigger(
            lambda: None
        )
        self.request_undo = self.input_generator.event_trigger(
            bpythonevents.UndoEvent
        )

        with self.input_generator:
            pass  # temp hack to get .original_stty

        super(FullCurtsiesRepl, self).__init__(
            locals_=locals_,
            config=config,
            banner=banner,
            interp=interp,
            orig_tcattrs=self.input_generator.original_stty,
        )

    def get_term_hw(self):
        return self.window.get_term_hw()

    def get_cursor_vertical_diff(self):
        return self.window.get_cursor_vertical_diff()

    def get_top_usable_line(self):
        return self.window.top_usable_row

    def on_suspend(self):
        self.window.__exit__(None, None, None)
        self.input_generator.__exit__(None, None, None)

    def after_suspend(self):
        self.input_generator.__enter__()
        self.window.__enter__()
        self.interrupting_refresh()

    def process_event_and_paint(self, e):
        """If None is passed in, just paint the screen"""
        try:
            if e is not None:
                self.process_event(e)
        except (SystemExitFromCodeRunner, SystemExit) as err:
            array, cursor_pos = self.paint(
                about_to_exit=True,
                user_quit=isinstance(err, SystemExitFromCodeRunner),
            )
            scrolled = self.window.render_to_terminal(array, cursor_pos)
            self.scroll_offset += scrolled
            raise
        else:
            array, cursor_pos = self.paint()
            scrolled = self.window.render_to_terminal(array, cursor_pos)
            self.scroll_offset += scrolled

    def mainloop(self, interactive=True, paste=None):
        if interactive:
            # Add custom help command
            # TODO: add methods to run the code
            self.initialize_interp()

            # run startup file
            self.process_event(bpythonevents.RunStartupFileEvent())

        # handle paste
        if paste:
            self.process_event(paste)

        # do a display before waiting for first event
        self.process_event_and_paint(None)
        inputs = combined_events(self.input_generator)
        for unused in find_iterator:
            e = inputs.send(0)
            if e is not None:
                self.process_event_and_paint(e)

        for e in inputs:
            self.process_event_and_paint(e)


def main(args=None, locals_=None, banner=None, welcome_message=None):
    """
    banner is displayed directly after the version information.
    welcome_message is passed on to Repl and displayed in the statusbar.
    """
    translations.init()

    config, options, exec_args = bpargs.parse(
        args,
        (
            "curtsies options",
            None,
            [
                Option(
                    "--log",
                    "-L",
                    action="count",
                    help=_("log debug messages to bpython.log"),
                ),
                Option(
                    "--paste",
                    "-p",
                    action="store_true",
                    help=_("start by pasting lines of a file into session"),
                ),
            ],
        ),
    )
    if options.log is None:
        options.log = 0
    logging_levels = [logging.ERROR, logging.INFO, logging.DEBUG]
    level = logging_levels[min(len(logging_levels) - 1, options.log)]
    logging.getLogger("curtsies").setLevel(level)
    logging.getLogger("bpython").setLevel(level)
    if options.log:
        handler = logging.FileHandler(filename="bpython.log")
        logging.getLogger("curtsies").addHandler(handler)
        logging.getLogger("curtsies").propagate = False
        logging.getLogger("bpython").addHandler(handler)
        logging.getLogger("bpython").propagate = False

    interp = None
    paste = None
    if exec_args:
        if not options:
            raise ValueError("don't pass in exec_args without options")
        exit_value = ()
        if options.paste:
            paste = curtsies.events.PasteEvent()
            encoding = inspection.get_encoding_file(exec_args[0])
            with io.open(exec_args[0], encoding=encoding) as f:
                sourcecode = f.read()
            paste.events.extend(sourcecode)
        else:
            try:
                interp = Interp(locals=locals_)
                bpargs.exec_code(interp, exec_args)
            except SystemExit as e:
                exit_value = e.args
            if not options.interactive:
                return extract_exit_value(exit_value)
    else:
        # expected for interactive sessions (vanilla python does it)
        sys.path.insert(0, "")

    if not options.quiet:
        print(bpargs.version_banner())
    if banner is not None:
        print(banner)

    if not py3:
        # XXX these deprecation warnings need to go at some point
        print(
            _(
                "WARNING: You are using `bpython` on Python 2. Support for Python 2 has been deprecated in version 0.19 and might disappear in a future version."
            )
        )

    global repl
    repl = FullCurtsiesRepl(config, locals_, welcome_message, interp)
    try:
        with repl.input_generator:
            with repl.window as win:
                with repl:
                    repl.height, repl.width = win.t.height, win.t.width
                    exit_value = repl.mainloop(True, paste)
    except (SystemExitFromCodeRunner, SystemExit) as e:
        exit_value = e.args
    return extract_exit_value(exit_value)


def _combined_events(event_provider, paste_threshold):
    """Combines consecutive keypress events into paste events."""
    timeout = yield "nonsense_event"  # so send can be used immediately
    queue = collections.deque()
    while True:
        e = event_provider.send(timeout)
        if isinstance(e, curtsies.events.Event):
            timeout = yield e
            continue
        elif e is None:
            timeout = yield None
            continue
        else:
            queue.append(e)
        e = event_provider.send(0)
        while not (e is None or isinstance(e, curtsies.events.Event)):
            queue.append(e)
            e = event_provider.send(0)
        if len(queue) >= paste_threshold:
            paste = curtsies.events.PasteEvent()
            paste.events.extend(queue)
            queue.clear()
            timeout = yield paste
        else:
            while len(queue):
                timeout = yield queue.popleft()


def combined_events(event_provider, paste_threshold=3):
    g = _combined_events(event_provider, paste_threshold)
    next(g)
    return g


if __name__ == "__main__":
    sys.exit(main())



# encoding: utf-8

import io
import os

from six.moves import range

from bpython.config import getpreferredencoding
from bpython.history import History
from bpython.test import unittest


class TestHistory(unittest.TestCase):
    def setUp(self):
        self.history = History("#%d" % x for x in range(1000))

    def test_is_at_start(self):
        self.history.first()

        self.assertNotEqual(self.history.index, 0)
        self.assertTrue(self.history.is_at_end)
        self.history.forward()
        self.assertFalse(self.history.is_at_end)

    def test_is_at_end(self):
        self.history.last()

        self.assertEqual(self.history.index, 0)
        self.assertTrue(self.history.is_at_start)
        self.assertFalse(self.history.is_at_end)

    def test_first(self):
        self.history.first()

        self.assertFalse(self.history.is_at_start)
        self.assertTrue(self.history.is_at_end)

    def test_last(self):
        self.history.last()

        self.assertTrue(self.history.is_at_start)
        self.assertFalse(self.history.is_at_end)

    def test_back(self):
        self.assertEqual(self.history.back(), "#999")
        self.assertNotEqual(self.history.back(), "#999")
        self.assertEqual(self.history.back(), "#997")
        for x in range(997):
            self.history.back()
        self.assertEqual(self.history.back(), "#0")

    def test_forward(self):
        self.history.first()

        self.assertEqual(self.history.forward(), "#1")
        self.assertNotEqual(self.history.forward(), "#1")
        self.assertEqual(self.history.forward(), "#3")
        #  1000 == entries   4 == len(range(1, 3) ===> '#1000' (so +1)
        for x in range(1000 - 4 - 1):
            self.history.forward()
        self.assertEqual(self.history.forward(), "#999")

    def test_append(self):
        self.history.append('print "foo\n"\n')
        self.history.append("\n")

        self.assertEqual(self.history.back(), 'print "foo\n"')

    def test_enter(self):
        self.history.enter("#lastnumber!")

        self.assertEqual(self.history.back(), "#lastnumber!")
        self.assertEqual(self.history.forward(), "#lastnumber!")

    def test_enter_2(self):
        self.history.enter("#50")

        self.assertEqual(self.history.back(), "#509")
        self.assertEqual(self.history.back(), "#508")
        self.assertEqual(self.history.forward(), "#509")

    def test_reset(self):
        self.history.enter("#lastnumber!")
        self.history.reset()

        self.assertEqual(self.history.back(), "#999")
        self.assertEqual(self.history.forward(), "")


class TestHistoryFileAccess(unittest.TestCase):
    def setUp(self):
        self.filename = "history_temp_file"
        self.encoding = getpreferredencoding()

        with io.open(
            self.filename, "w", encoding=self.encoding, errors="ignore"
        ) as f:
            f.write(b"#1\n#2\n".decode())

    def test_load(self):
        history = History()

        history.load(self.filename, self.encoding)
        self.assertEqual(history.entries, ["#1", "#2"])

    def test_append_reload_and_write(self):
        history = History()

        history.append_reload_and_write("#3", self.filename, self.encoding)
        self.assertEqual(history.entries, ["#1", "#2", "#3"])

        history.append_reload_and_write("#4", self.filename, self.encoding)
        self.assertEqual(history.entries, ["#1", "#2", "#3", "#4"])

    def test_save(self):
        history = History()
        history.entries = []
        for line in ["#1", "#2", "#3", "#4"]:
            history.append_to(history.entries, line)

        # save only last 2 lines
        history.save(self.filename, self.encoding, lines=2)

        # empty the list of entries and load again from the file
        history.entries = [""]
        history.load(self.filename, self.encoding)

        self.assertEqual(history.entries, ["#3", "#4"])

    def tearDown(self):
        try:
            os.remove(self.filename)
        except OSError:
            pass



# encoding: utf-8

from __future__ import unicode_literals

from bpython import importcompletion
from bpython.test import unittest


class TestSimpleComplete(unittest.TestCase):
    def setUp(self):
        self.original_modules = importcompletion.modules
        importcompletion.modules = [
            "zzabc",
            "zzabd",
            "zzefg",
            "zzabc.e",
            "zzabc.f",
        ]

    def tearDown(self):
        importcompletion.modules = self.original_modules

    def test_simple_completion(self):
        self.assertSetEqual(
            importcompletion.complete(10, "import zza"), set(["zzabc", "zzabd"])
        )

    def test_package_completion(self):
        self.assertSetEqual(
            importcompletion.complete(13, "import zzabc."),
            set(["zzabc.e", "zzabc.f"]),
        )


class TestRealComplete(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        for _ in importcompletion.find_iterator:
            pass
        __import__("sys")
        __import__("os")

    @classmethod
    def tearDownClass(cls):
        importcompletion.find_iterator = importcompletion.find_all_modules()
        importcompletion.modules = set()

    def test_from_attribute(self):
        self.assertSetEqual(
            importcompletion.complete(19, "from sys import arg"), set(["argv"])
        )

    def test_from_attr_module(self):
        self.assertSetEqual(
            importcompletion.complete(9, "from os.p"), set(["os.path"])
        )

    def test_from_package(self):
        self.assertSetEqual(
            importcompletion.complete(17, "from xml import d"), set(["dom"])
        )



# encoding: utf-8

"""Non-keyboard events used in bpython curtsies REPL"""
import time

import curtsies.events


class ReloadEvent(curtsies.events.Event):
    """Request to rerun REPL session ASAP because imported modules changed"""

    def __init__(self, files_modified=("?",)):
        self.files_modified = files_modified

    def __repr__(self):
        return "<ReloadEvent from %s>" % (" & ".join(self.files_modified))


class RefreshRequestEvent(curtsies.events.Event):
    """Request to refresh REPL display ASAP"""

    def __repr__(self):
        return "<RefreshRequestEvent for now>"


class ScheduledRefreshRequestEvent(curtsies.events.ScheduledEvent):
    """Request to refresh the REPL display at some point in the future

    Used to schedule the disappearance of status bar message that only shows
    for a few seconds"""

    def __init__(self, when):
        super(ScheduledRefreshRequestEvent, self).__init__(when)

    def __repr__(self):
        return "<RefreshRequestEvent for %s seconds from now>" % (
            self.when - time.time()
        )


class RunStartupFileEvent(curtsies.events.Event):
    """Request to run the startup file."""


class UndoEvent(curtsies.events.Event):
    """Request to undo."""

    def __init__(self, n=1):
        self.n = n



print("import: 'bpdb'")
import bpdb

print("import: 'bpython'")
import bpython

print("import: 'bpython.curtsiesfrontend'")
import bpython.curtsiesfrontend

print("import: 'bpython.test'")
import bpython.test

print("import: 'bpython.test.fodder'")
import bpython.test.fodder

print("import: 'bpython.translations'")
import bpython.translations




# encoding: utf-8

import os
import sys
import tempfile

from bpython._py3compat import py3
from bpython.test import unittest
from bpython.importcompletion import find_modules


@unittest.skipIf(not py3, "Test doesn't work in python 2.")
class TestAvoidSymbolicLinks(unittest.TestCase):
    def setUp(self):
        with tempfile.TemporaryDirectory() as import_test_folder:
            os.mkdir(os.path.join(import_test_folder, "Level0"))
            os.mkdir(os.path.join(import_test_folder, "Right"))
            os.mkdir(os.path.join(import_test_folder, "Left"))

            current_path = os.path.join(import_test_folder, "Level0")
            with open(
                os.path.join(current_path, "__init__.py"), "x"
            ) as init_file:
                pass

            current_path = os.path.join(current_path, "Level1")
            os.mkdir(current_path)
            with open(
                os.path.join(current_path, "__init__.py"), "x"
            ) as init_file:
                pass

            current_path = os.path.join(current_path, "Level2")
            os.mkdir(current_path)
            with open(
                os.path.join(current_path, "__init__.py"), "x"
            ) as init_file:
                pass

            os.symlink(
                os.path.join(import_test_folder, "Level0/Level1"),
                os.path.join(current_path, "Level3"),
                True,
            )

            current_path = os.path.join(import_test_folder, "Right")
            with open(
                os.path.join(current_path, "__init__.py"), "x"
            ) as init_file:
                pass

            os.symlink(
                os.path.join(import_test_folder, "Left"),
                os.path.join(current_path, "toLeft"),
                True,
            )

            current_path = os.path.join(import_test_folder, "Left")
            with open(
                os.path.join(current_path, "__init__.py"), "x"
            ) as init_file:
                pass

            os.symlink(
                os.path.join(import_test_folder, "Right"),
                os.path.join(current_path, "toRight"),
                True,
            )

            self.modules = list(
                find_modules(os.path.abspath(import_test_folder))
            )
            self.filepaths = [
                "Left.toRight.toLeft",
                "Left.toRight",
                "Left",
                "Level0.Level1.Level2.Level3",
                "Level0.Level1.Level2",
                "Level0.Level1",
                "Level0",
                "Right",
                "Right.toLeft",
                "Right.toLeft.toRight",
            ]

    def test_simple_symbolic_link_loop(self):
        for thing in self.modules:
            self.assertTrue(thing in self.filepaths)
            if thing == "Left.toRight.toLeft":
                self.filepaths.remove("Right.toLeft")
                self.filepaths.remove("Right.toLeft.toRight")
            if thing == "Right.toLeft.toRight":
                self.filepaths.remove("Left.toRight.toLeft")
                self.filepaths.remove("Left.toRight")
            self.filepaths.remove(thing)
        self.assertFalse(self.filepaths)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
from __future__ import unicode_literals

from functools import partial
import re

from bpython.lazyre import LazyReCompile

from curtsies.termformatconstants import FG_COLORS, BG_COLORS, colors
from curtsies.formatstring import fmtstr, FmtStr


cnames = dict(zip("krgybmcwd", colors + ("default",)))


def func_for_letter(l, default="k"):
    """Returns FmtStr constructor for a bpython-style color code"""
    if l == "d":
        l = default
    elif l == "D":
        l = default.upper()
    return partial(fmtstr, fg=cnames[l.lower()], bold=l.isupper())


def color_for_letter(l, default="k"):
    if l == "d":
        l = default
    return cnames[l.lower()]


def parse(s):
    """Returns a FmtStr object from a bpython-formatted colored string"""
    rest = s
    stuff = []
    while True:
        if not rest:
            break
        start, rest = peel_off_string(rest)
        stuff.append(start)
    return (
        sum((fs_from_match(d) for d in stuff[1:]), fs_from_match(stuff[0]))
        if len(stuff) > 0
        else FmtStr()
    )


def fs_from_match(d):
    atts = {}
    if d["fg"]:

        # this isn't according to spec as I understand it
        if d["fg"].isupper():
            d["bold"] = True
        # TODO figure out why boldness isn't based on presence of \x02

        color = cnames[d["fg"].lower()]
        if color != "default":
            atts["fg"] = FG_COLORS[color]
    if d["bg"]:
        if d["bg"] == "I":
            # hack for finding the "inverse"
            color = colors[
                (colors.index(color) + (len(colors) // 2)) % len(colors)
            ]
        else:
            color = cnames[d["bg"].lower()]
        if color != "default":
            atts["bg"] = BG_COLORS[color]
    if d["bold"]:
        atts["bold"] = True
    return fmtstr(d["string"], **atts)


peel_off_string_re = LazyReCompile(
    r"""(?P<colormarker>\x01
            (?P<fg>[krgybmcwdKRGYBMCWD]?)
            (?P<bg>[krgybmcwdKRGYBMCWDI]?)?)
        (?P<bold>\x02?)
        \x03
        (?P<string>[^\x04]*)
        \x04
        (?P<rest>.*)
        """,
    re.VERBOSE | re.DOTALL,
)


def peel_off_string(s):
    m = peel_off_string_re.match(s)
    assert m, repr(s)
    d = m.groupdict()
    rest = d["rest"]
    del d["rest"]
    return d, rest



# encoding: utf-8

from code import compile_command as compiler
from functools import partial
import difflib
import inspect
import re

from bpython.curtsiesfrontend.interpreter import code_finished_will_parse
from bpython.curtsiesfrontend.preprocess import preprocess
from bpython.test import unittest
from bpython.test.fodder import original, processed


skip = unittest.skip
preproc = partial(preprocess, compiler=compiler)


def get_fodder_source(test_name):
    pattern = r"#StartTest-%s\n(.*?)#EndTest" % (test_name,)
    orig, xformed = [
        re.search(pattern, inspect.getsource(module), re.DOTALL)
        for module in [original, processed]
    ]

    if not orig:
        raise ValueError(
            "Can't locate test %s in original fodder file" % (test_name,)
        )
    if not xformed:
        raise ValueError(
            "Can't locate test %s in processed fodder file" % (test_name,)
        )
    return orig.group(1), xformed.group(1)


class TestPreprocessing(unittest.TestCase):
    def assertCompiles(self, source):
        finished, parsable = code_finished_will_parse(source, compiler)
        return finished and parsable

    def test_indent_empty_lines_nops(self):
        self.assertEqual(preproc("hello"), "hello")
        self.assertEqual(preproc("hello\ngoodbye"), "hello\ngoodbye")
        self.assertEqual(preproc("a\n    b\nc\n"), "a\n    b\nc\n")

    def assertShowWhitespaceEqual(self, a, b):
        self.assertEqual(
            a,
            b,
            "".join(
                difflib.context_diff(
                    a.replace(" ", "~").splitlines(True),
                    b.replace(" ", "~").splitlines(True),
                    fromfile="actual",
                    tofile="expected",
                    n=5,
                )
            ),
        )

    def assertDefinitionIndented(self, obj):
        name = obj.__name__
        obj2 = getattr(processed, name)
        orig = inspect.getsource(obj)
        xformed = inspect.getsource(obj2)
        self.assertShowWhitespaceEqual(preproc(orig), xformed)
        self.assertCompiles(xformed)

    def assertLinesIndented(self, test_name):
        orig, xformed = get_fodder_source(test_name)
        self.assertShowWhitespaceEqual(preproc(orig), xformed)
        self.assertCompiles(xformed)

    def assertIndented(self, obj_or_name):
        if isinstance(obj_or_name, str):
            self.assertLinesIndented(obj_or_name)
        else:
            self.assertDefinitionIndented(obj_or_name)

    def test_empty_line_between_methods(self):
        self.assertIndented(original.BlankLineBetweenMethods)

    def test_empty_line_within_class(self):
        self.assertIndented(original.BlankLineInFunction)

    def test_blank_lines_in_for_loop(self):
        self.assertIndented("blank_lines_in_for_loop")

    @skip(
        "More advanced technique required: need to try compiling and "
        "backtracking"
    )
    def test_blank_line_in_try_catch(self):
        self.assertIndented("blank_line_in_try_catch")

    @skip(
        "More advanced technique required: need to try compiling and "
        "backtracking"
    )
    def test_blank_line_in_try_catch_else(self):
        self.assertIndented("blank_line_in_try_catch_else")

    def test_blank_trailing_line(self):
        self.assertIndented("blank_trailing_line")

    def test_tabs(self):
        self.assertIndented(original.tabs)



# encoding: utf-8

"""For running Python code that could interrupt itself at any time in order to,
for example, ask for a read on stdin, or a write on stdout

The CodeRunner spawns a greenlet to run code in, and that code can suspend its
own execution to ask the main greenlet to refresh the display or get
information.

Greenlets are basically threads that can explicitly switch control to each
other.  You can replace the word "greenlet" with "thread" in these docs if that
makes more sense to you.
"""

import code
import signal
import greenlet
import logging

from bpython._py3compat import py3, is_main_thread
from bpython.config import getpreferredencoding

logger = logging.getLogger(__name__)


class SigintHappened(object):
    """If this class is returned, a SIGINT happened while the main greenlet"""


class SystemExitFromCodeRunner(SystemExit):
    """If this class is returned, a SystemExit happened while in the code
    greenlet"""


class RequestFromCodeRunner(object):
    """Message from the code runner"""


class Wait(RequestFromCodeRunner):
    """Running code would like the main loop to run for a bit"""


class Refresh(RequestFromCodeRunner):
    """Running code would like the main loop to refresh the display"""


class Done(RequestFromCodeRunner):
    """Running code is done running"""


class Unfinished(RequestFromCodeRunner):
    """Source code wasn't executed because it wasn't fully formed"""


class SystemExitRequest(RequestFromCodeRunner):
    """Running code raised a SystemExit"""

    def __init__(self, args):
        self.args = args


class CodeRunner(object):
    """Runs user code in an interpreter.

    Running code requests a refresh by calling
    request_from_main_context(force_refresh=True), which
    suspends execution of the code and switches back to the main greenlet

    After load_code() is called with the source code to be run,
    the run_code() method should be called to start running the code.
    The running code may request screen refreshes and user input
    by calling request_from_main_context.
    When this are called, the running source code cedes
    control, and the current run_code() method call returns.

    The return value of run_code() determines whether the method ought
    to be called again to complete execution of the source code.

    Once the screen refresh has occurred or the requested user input
    has been gathered, run_code() should be called again, passing in any
    requested user input. This continues until run_code returns Done.

    The code greenlet is responsible for telling the main greenlet
    what it wants returned in the next run_code call - CodeRunner
    just passes whatever is passed in to run_code(for_code) to the
    code greenlet
    """

    def __init__(self, interp=None, request_refresh=lambda: None):
        """
        interp is an interpreter object to use. By default a new one is
        created.

        request_refresh is a function that will be called each time the running
        code asks for a refresh - to, for example, update the screen.
        """
        self.interp = interp or code.InteractiveInterpreter()
        self.source = None
        self.main_context = greenlet.getcurrent()
        self.code_context = None
        self.request_refresh = request_refresh
        # waiting for response from main thread
        self.code_is_waiting = False
        # sigint happened while in main thread
        self.sigint_happened_in_main_context = False
        self.orig_sigint_handler = None

    @property
    def running(self):
        """Returns greenlet if code has been loaded greenlet has been
        started"""
        return self.source and self.code_context

    def load_code(self, source):
        """Prep code to be run"""
        assert self.source is None, (
            "you shouldn't load code when some is " "already running"
        )
        self.source = source
        self.code_context = None

    def _unload_code(self):
        """Called when done running code"""
        self.source = None
        self.code_context = None
        self.code_is_waiting = False

    def run_code(self, for_code=None):
        """Returns Truthy values if code finishes, False otherwise

        if for_code is provided, send that value to the code greenlet
        if source code is complete, returns "done"
        if source code is incomplete, returns "unfinished"
        """
        if self.code_context is None:
            assert self.source is not None
            self.code_context = greenlet.greenlet(self._blocking_run_code)
            if is_main_thread():
                self.orig_sigint_handler = signal.getsignal(signal.SIGINT)
                signal.signal(signal.SIGINT, self.sigint_handler)
            request = self.code_context.switch()
        else:
            assert self.code_is_waiting
            self.code_is_waiting = False
            if is_main_thread():
                signal.signal(signal.SIGINT, self.sigint_handler)
            if self.sigint_happened_in_main_context:
                self.sigint_happened_in_main_context = False
                request = self.code_context.switch(SigintHappened)
            else:
                request = self.code_context.switch(for_code)

        logger.debug("request received from code was %r", request)
        if not isinstance(request, RequestFromCodeRunner):
            raise ValueError(
                "Not a valid value from code greenlet: %r" % request
            )
        if isinstance(request, (Wait, Refresh)):
            self.code_is_waiting = True
            if isinstance(request, Refresh):
                self.request_refresh()
            return False
        elif isinstance(request, (Done, Unfinished)):
            self._unload_code()
            if is_main_thread():
                signal.signal(signal.SIGINT, self.orig_sigint_handler)
            self.orig_sigint_handler = None
            return request
        elif isinstance(request, SystemExitRequest):
            self._unload_code()
            raise SystemExitFromCodeRunner(request.args)

    def sigint_handler(self, *args):
        """SIGINT handler to use while code is running or request being
        fulfilled"""
        if greenlet.getcurrent() is self.code_context:
            logger.debug("sigint while running user code!")
            raise KeyboardInterrupt()
        else:
            logger.debug(
                "sigint while fulfilling code request sigint handler "
                "running!"
            )
            self.sigint_happened_in_main_context = True

    def _blocking_run_code(self):
        try:
            unfinished = self.interp.runsource(self.source)
        except SystemExit as e:
            return SystemExitRequest(*e.args)
        return Unfinished() if unfinished else Done()

    def request_from_main_context(self, force_refresh=False):
        """Return the argument passed in to .run_code(for_code)

        Nothing means calls to run_code must be... ???
        """
        if force_refresh:
            value = self.main_context.switch(Refresh())
        else:
            value = self.main_context.switch(Wait())
        if value is SigintHappened:
            raise KeyboardInterrupt()
        return value


class FakeOutput(object):
    def __init__(self, coderunner, on_write, real_fileobj):
        """Fakes sys.stdout or sys.stderr

        on_write should always take unicode

        fileno should be the fileno that on_write will
                output to (e.g. 1 for standard output).
        """
        self.coderunner = coderunner
        self.on_write = on_write
        self._real_fileobj = real_fileobj

    def write(self, s, *args, **kwargs):
        if not py3 and isinstance(s, str):
            s = s.decode(getpreferredencoding(), "ignore")
        self.on_write(s, *args, **kwargs)
        return self.coderunner.request_from_main_context(force_refresh=True)

    # Some applications which use curses require that sys.stdout
    # have a method called fileno. One example is pwntools. This
    # is not a widespread issue, but is annoying.
    def fileno(self):
        return self._real_fileobj.fileno()

    def writelines(self, l):
        for s in l:
            self.write(s)

    def flush(self):
        pass

    def isatty(self):
        return True

    @property
    def encoding(self):
        return self._real_fileobj.encoding



# encoding: utf-8

# The MIT License
#
# Copyright (c) 2009-2011 Andreas Stuehrk
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from ._py3compat import py3, try_decode
from .line import (
    current_word,
    current_import,
    current_from_import_from,
    current_from_import_import,
)

import fnmatch
import os
import sys
import warnings
from six.moves import filter

if py3:
    import importlib.machinery

    SUFFIXES = importlib.machinery.all_suffixes()
else:
    import imp

    SUFFIXES = [suffix for suffix, mode, type in imp.get_suffixes()]

# The cached list of all known modules
modules = set()
# List of stored paths to compare against so that real paths are not repeated
# handles symlinks not mount points
paths = set()
# Patterns to skip
# TODO: This skiplist should be configurable.
skiplist = (".git", ".config", ".local", ".share", "node_modules")
fully_loaded = False


def module_matches(cw, prefix=""):
    """Modules names to replace cw with"""
    full = "%s.%s" % (prefix, cw) if prefix else cw
    matches = (
        name
        for name in modules
        if (name.startswith(full) and name.find(".", len(full)) == -1)
    )
    if prefix:
        return set(match[len(prefix) + 1 :] for match in matches)
    else:
        return set(matches)


def attr_matches(cw, prefix="", only_modules=False):
    """Attributes to replace name with"""
    full = "%s.%s" % (prefix, cw) if prefix else cw
    module_name, _, name_after_dot = full.rpartition(".")
    if module_name not in sys.modules:
        return set()
    module = sys.modules[module_name]
    if only_modules:
        matches = (
            name
            for name in dir(module)
            if (
                name.startswith(name_after_dot)
                and "%s.%s" % (module_name, name)
            )
            in sys.modules
        )
    else:
        matches = (
            name for name in dir(module) if name.startswith(name_after_dot)
        )
    module_part, _, _ = cw.rpartition(".")
    if module_part:
        matches = ("%s.%s" % (module_part, m) for m in matches)

    generator = (try_decode(match, "ascii") for match in matches)
    return set(filter(lambda x: x is not None, generator))


def module_attr_matches(name):
    """Only attributes which are modules to replace name with"""
    return attr_matches(name, prefix="", only_modules=True)


def complete(cursor_offset, line):
    """Construct a full list of possibly completions for imports."""
    tokens = line.split()
    if "from" not in tokens and "import" not in tokens:
        return None

    result = current_word(cursor_offset, line)
    if result is None:
        return None

    from_import_from = current_from_import_from(cursor_offset, line)
    if from_import_from is not None:
        import_import = current_from_import_import(cursor_offset, line)
        if import_import is not None:
            # `from a import <b|>` completion
            matches = module_matches(import_import[2], from_import_from[2])
            matches.update(attr_matches(import_import[2], from_import_from[2]))
        else:
            # `from <a|>` completion
            matches = module_attr_matches(from_import_from[2])
            matches.update(module_matches(from_import_from[2]))
        return matches

    cur_import = current_import(cursor_offset, line)
    if cur_import is not None:
        # `import <a|>` completion
        matches = module_matches(cur_import[2])
        matches.update(module_attr_matches(cur_import[2]))
        return matches
    else:
        return None


def find_modules(path):
    """Find all modules (and packages) for a given directory."""
    if not os.path.isdir(path):
        # Perhaps a zip file
        return
    basepath = os.path.basename(path)
    if any(fnmatch.fnmatch(basepath, entry) for entry in skiplist):
        # Path is on skiplist
        return

    try:
        filenames = os.listdir(path)
    except EnvironmentError:
        filenames = []

    if py3:
        finder = importlib.machinery.FileFinder(path)

    for name in filenames:
        if any(fnmatch.fnmatch(name, entry) for entry in skiplist):
            # Path is on skiplist
            continue
        elif not any(name.endswith(suffix) for suffix in SUFFIXES):
            # Possibly a package
            if "." in name:
                continue
        elif os.path.isdir(os.path.join(path, name)):
            # Unfortunately, CPython just crashes if there is a directory
            # which ends with a python extension, so work around.
            continue
        for suffix in SUFFIXES:
            if name.endswith(suffix):
                name = name[: -len(suffix)]
                break
        if py3 and name == "badsyntax_pep3120":
            # Workaround for issue #166
            continue
        try:
            is_package = False
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", ImportWarning)
                if py3:
                    spec = finder.find_spec(name)
                    if spec is None:
                        continue
                    if spec.submodule_search_locations is not None:
                        pathname = spec.submodule_search_locations[0]
                        is_package = True
                    else:
                        pathname = spec.origin
                else:
                    fo, pathname, _ = imp.find_module(name, [path])
                    if fo is not None:
                        fo.close()
                    else:
                        # Yay, package
                        is_package = True
        except (ImportError, IOError, SyntaxError):
            continue
        except UnicodeEncodeError:
            # Happens with Python 3 when there is a filename in some
            # invalid encoding
            continue
        else:
            if is_package:
                path_real = os.path.realpath(pathname)
                if path_real not in paths:
                    paths.add(path_real)
                    for subname in find_modules(pathname):
                        if subname != "__init__":
                            yield "%s.%s" % (name, subname)
            yield name


def find_all_modules(path=None):
    """Return a list with all modules in `path`, which should be a list of
    directory names. If path is not given, sys.path will be used."""
    if path is None:
        modules.update(try_decode(m, "ascii") for m in sys.builtin_module_names)
        path = sys.path

    for p in path:
        if not p:
            p = os.curdir
        for module in find_modules(p):
            module = try_decode(module, "ascii")
            if module is None:
                continue
            modules.add(module)
            yield


def find_coroutine():
    global fully_loaded

    if fully_loaded:
        return None

    try:
        next(find_iterator)
    except StopIteration:
        fully_loaded = True

    return True


def reload():
    """Refresh the list of known modules."""
    modules.clear()
    for _ in find_all_modules():
        pass


find_iterator = find_all_modules()



# -*- coding: utf-8 -*-

import ast
import numbers
import sys

from bpython.simpleeval import (
    simple_eval,
    evaluate_current_expression,
    EvaluationError,
)
from bpython.test import unittest
from bpython._py3compat import py3


class TestSimpleEval(unittest.TestCase):
    def assertMatchesStdlib(self, expr):
        self.assertEqual(ast.literal_eval(expr), simple_eval(expr))

    def test_matches_stdlib(self):
        """Should match the stdlib literal_eval if no names or indexing"""
        self.assertMatchesStdlib("[1]")
        self.assertMatchesStdlib("{(1,): [2,3,{}]}")

    @unittest.skipUnless(py3, "Only Python3 versions of ast.literal_eval evaluate set literals")
    def test_matches_stdlib_py3(self):
        """Should match the stdlib literal_eval if no names or indexing"""
        self.assertMatchesStdlib("{1, 2}")

    @unittest.skipUnless(sys.version_info[:2] >= (3, 9), "Only Python3.9 evaluates set()")
    def test_matches_stdlib_set_literal(self):
        """set() is evaluated"""
        self.assertMatchesStdlib("set()")

    def test_indexing(self):
        """Literals can be indexed into"""
        self.assertEqual(simple_eval("[1,2][0]"), 1)

    def test_name_lookup(self):
        """Names can be looked up in a namespace"""
        self.assertEqual(simple_eval("a", {"a": 1}), 1)
        self.assertEqual(simple_eval("map"), map)

    def test_name_lookup_indexing(self):
        """Names can be looked up in a namespace"""
        self.assertEqual(simple_eval("a[b]", {"a": {"c": 1}, "b": "c"}), 1)

    def test_lookup_on_suspicious_types(self):
        class FakeDict(object):
            pass

        with self.assertRaises(ValueError):
            simple_eval("a[1]", {"a": FakeDict()})

        class TrickyDict(dict):
            def __getitem__(self, index):
                self.fail("doing key lookup isn't safe")

        with self.assertRaises(ValueError):
            simple_eval("a[1]", {"a": TrickyDict()})

        class SchrodingersDict(dict):
            def __getattribute__(inner_self, attr):
                self.fail("doing attribute lookup might have side effects")

        with self.assertRaises(ValueError):
            simple_eval("a[1]", {"a": SchrodingersDict()})

        class SchrodingersCatsDict(dict):
            def __getattr__(inner_self, attr):
                self.fail("doing attribute lookup might have side effects")

        with self.assertRaises(ValueError):
            simple_eval("a[1]", {"a": SchrodingersDict()})

    def test_operators_on_suspicious_types(self):
        class Spam(numbers.Number):
            def __add__(inner_self, other):
                self.fail("doing attribute lookup might have side effects")

        with self.assertRaises(ValueError):
            simple_eval("a + 1", {"a": Spam()})

    def test_operators_on_numbers(self):
        self.assertEqual(simple_eval("-2"), -2)
        self.assertEqual(simple_eval("1 + 1"), 2)
        self.assertEqual(simple_eval("a - 2", {"a": 1}), -1)
        with self.assertRaises(ValueError):
            simple_eval("2 * 3")
        with self.assertRaises(ValueError):
            simple_eval("2 ** 3")

    def test_function_calls_raise(self):
        with self.assertRaises(ValueError):
            simple_eval("1()")

    def test_nonexistant_names_raise(self):
        with self.assertRaises(EvaluationError):
            simple_eval("a")

    def test_attribute_access(self):
        class Foo(object):
            abc = 1

        self.assertEqual(simple_eval("foo.abc", {"foo": Foo()}), 1)


class TestEvaluateCurrentExpression(unittest.TestCase):
    def assertEvaled(self, line, value, ns=None):
        assert line.count("|") == 1
        cursor_offset = line.find("|")
        line = line.replace("|", "")
        self.assertEqual(
            evaluate_current_expression(cursor_offset, line, ns), value
        )

    def assertCannotEval(self, line, ns=None):
        assert line.count("|") == 1
        cursor_offset = line.find("|")
        line = line.replace("|", "")
        with self.assertRaises(EvaluationError):
            evaluate_current_expression(cursor_offset, line, ns)

    def test_simple(self):
        self.assertEvaled("[1].a|bc", [1])
        self.assertEvaled("[1].abc|", [1])
        self.assertEvaled("[1].|abc", [1])
        self.assertEvaled("[1]. |abc", [1])
        self.assertEvaled("[1] .|abc", [1])
        self.assertCannotEval("[1].abc |", [1])
        self.assertCannotEval("[1]. abc |", [1])
        self.assertCannotEval("[2][1].a|bc", [1])

    def test_nonsense(self):
        self.assertEvaled("!@#$ [1].a|bc", [1])
        self.assertEvaled("--- [2][0].a|bc", 2)
        self.assertCannotEval('"asdf".centered()[1].a|bc')
        self.assertEvaled('"asdf"[1].a|bc', "s")

    def test_with_namespace(self):
        self.assertEvaled("a[1].a|bc", "d", {"a": "adsf"})
        self.assertCannotEval("a[1].a|bc", {})


if __name__ == "__main__":
    unittest.main()



# coding: utf8

from __future__ import unicode_literals
import itertools
import os
import pydoc
import string
import sys
from contextlib import contextmanager

from curtsies.formatstringarray import FormatStringTest, fsarray
from curtsies.fmtfuncs import cyan, bold, green, yellow, on_magenta, red

from bpython.curtsiesfrontend.events import RefreshRequestEvent
from bpython.test import mock
from bpython import config, inspection
from bpython.curtsiesfrontend.repl import BaseRepl
from bpython.curtsiesfrontend import replpainter
from bpython.curtsiesfrontend.repl import (
    INCONSISTENT_HISTORY_MSG,
    CONTIGUITY_BROKEN_MSG,
)
from bpython.test import FixLanguageTestCase as TestCase, TEST_CONFIG


def setup_config():
    config_struct = config.Struct()
    config.loadini(config_struct, TEST_CONFIG)
    config_struct.cli_suggestion_width = 1
    return config_struct


class ClearEnviron(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mock_environ = mock.patch.dict("os.environ", {}, clear=True)
        cls.mock_environ.start()
        TestCase.setUpClass()

    @classmethod
    def tearDownClass(cls):
        cls.mock_environ.stop()
        TestCase.tearDownClass()


class CurtsiesPaintingTest(FormatStringTest, ClearEnviron):
    def setUp(self):
        class TestRepl(BaseRepl):
            def _request_refresh(inner_self):
                pass

        self.repl = TestRepl(config=setup_config())
        self.repl.height, self.repl.width = (5, 10)

    @property
    def locals(self):
        return self.repl.coderunner.interp.locals

    def assert_paint(self, screen, cursor_row_col):
        array, cursor_pos = self.repl.paint()
        self.assertFSArraysEqual(array, screen)
        self.assertEqual(cursor_pos, cursor_row_col)

    def assert_paint_ignoring_formatting(
        self, screen, cursor_row_col=None, **paint_kwargs
    ):
        array, cursor_pos = self.repl.paint(**paint_kwargs)
        self.assertFSArraysEqualIgnoringFormatting(array, screen)
        if cursor_row_col is not None:
            self.assertEqual(cursor_pos, cursor_row_col)


class TestCurtsiesPaintingTest(CurtsiesPaintingTest):
    def test_history_is_cleared(self):
        self.assertEqual(self.repl.rl_history.entries, [""])


class TestCurtsiesPaintingSimple(CurtsiesPaintingTest):
    def test_startup(self):
        screen = fsarray([cyan(">>> "), cyan("Welcome to")])
        self.assert_paint(screen, (0, 4))

    def test_enter_text(self):
        [self.repl.add_normal_character(c) for c in "1 + 1"]
        screen = fsarray(
            [
                cyan(">>> ")
                + bold(
                    green("1")
                    + cyan(" ")
                    + yellow("+")
                    + cyan(" ")
                    + green("1")
                ),
                cyan("Welcome to"),
            ]
        )
        self.assert_paint(screen, (0, 9))

    def test_run_line(self):
        try:
            orig_stdout = sys.stdout
            sys.stdout = self.repl.stdout
            [self.repl.add_normal_character(c) for c in "1 + 1"]
            self.repl.on_enter(new_code=False)
            screen = fsarray([">>> 1 + 1", "2", "Welcome to"])
            self.assert_paint_ignoring_formatting(screen, (1, 1))
        finally:
            sys.stdout = orig_stdout

    def test_completion(self):
        self.repl.height, self.repl.width = (5, 32)
        self.repl.current_line = "an"
        self.cursor_offset = 2
        if config.supports_box_chars():
            screen = [
                ">>> an",
                "┌──────────────────────────────┐",
                "│ and  any(                    │",
                "└──────────────────────────────┘",
                "Welcome to bpython! Press <F1> f",
            ]
        else:
            screen = [
                ">>> an",
                "+------------------------------+",
                "| and  any(                    |",
                "+------------------------------+",
                "Welcome to bpython! Press <F1> f",
            ]
        self.assert_paint_ignoring_formatting(screen, (0, 4))

    def test_argspec(self):
        def foo(x, y, z=10):
            "docstring!"
            pass

        argspec = inspection.getfuncprops("foo", foo)
        array = replpainter.formatted_argspec(argspec, 1, 30, setup_config())
        screen = [
            bold(cyan("foo"))
            + cyan(":")
            + cyan(" ")
            + cyan("(")
            + cyan("x")
            + yellow(",")
            + yellow(" ")
            + bold(cyan("y"))
            + yellow(",")
            + yellow(" ")
            + cyan("z")
            + yellow("=")
            + bold(cyan("10"))
            + yellow(")")
        ]
        self.assertFSArraysEqual(fsarray(array), fsarray(screen))

    def test_formatted_docstring(self):
        actual = replpainter.formatted_docstring(
            "Returns the results\n\n" "Also has side effects",
            40,
            config=setup_config(),
        )
        expected = fsarray(["Returns the results", "", "Also has side effects"])
        self.assertFSArraysEqualIgnoringFormatting(actual, expected)

    def test_unicode_docstrings(self):
        "A bit of a special case in Python 2"
        # issue 653

        def foo():
            "åß∂ƒ"

        actual = replpainter.formatted_docstring(
            foo.__doc__, 40, config=setup_config()
        )
        expected = fsarray(["åß∂ƒ"])
        self.assertFSArraysEqualIgnoringFormatting(actual, expected)

    def test_nonsense_docstrings(self):
        for docstring in [
            123,
            {},
            [],
        ]:
            try:
                replpainter.formatted_docstring(
                    docstring, 40, config=setup_config()
                )
            except Exception:
                self.fail("bad docstring caused crash: {!r}".format(docstring))

    def test_weird_boto_docstrings(self):
        # Boto does something like this.
        # botocore: botocore/docs/docstring.py
        class WeirdDocstring(str):
            # a mighty hack. See botocore/docs/docstring.py
            def expandtabs(self, tabsize=8):
                return "asdfåß∂ƒ".expandtabs(tabsize)

        def foo():
            pass

        foo.__doc__ = WeirdDocstring()
        wd = pydoc.getdoc(foo)
        actual = replpainter.formatted_docstring(wd, 40, config=setup_config())
        expected = fsarray(["asdfåß∂ƒ"])
        self.assertFSArraysEqualIgnoringFormatting(actual, expected)

    def test_paint_lasts_events(self):
        actual = replpainter.paint_last_events(
            4, 100, ["a", "b", "c"], config=setup_config()
        )
        if config.supports_box_chars():
            expected = fsarray(["┌─┐", "│c│", "│b│", "└─┘"])
        else:
            expected = fsarray(["+-+", "|c|", "|b|", "+-+"])

        self.assertFSArraysEqualIgnoringFormatting(actual, expected)


@contextmanager
def output_to_repl(repl):
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = repl.stdout, repl.stderr
        yield
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class HigherLevelCurtsiesPaintingTest(CurtsiesPaintingTest):
    def refresh(self):
        self.refresh_requests.append(RefreshRequestEvent())

    def send_refreshes(self):
        while self.refresh_requests:
            self.repl.process_event(self.refresh_requests.pop())
            _, _ = self.repl.paint()

    def enter(self, line=None):
        """Enter a line of text, avoiding autocompletion windows

        autocomplete could still happen if the entered line has
        autocompletion that would happen then, but intermediate
        stages won't happen"""
        if line is not None:
            self.repl._set_cursor_offset(len(line), update_completion=False)
            self.repl.current_line = line
        with output_to_repl(self.repl):
            self.repl.on_enter(new_code=False)
            self.assertEqual(self.repl.rl_history.entries, [""])
            self.send_refreshes()

    def undo(self):
        with output_to_repl(self.repl):
            self.repl.undo()
            self.send_refreshes()

    def setUp(self):
        self.refresh_requests = []

        class TestRepl(BaseRepl):
            def _request_refresh(inner_self):
                self.refresh()

        self.repl = TestRepl(banner="", config=setup_config())
        self.repl.height, self.repl.width = (5, 32)

    def send_key(self, key):
        self.repl.process_event("<SPACE>" if key == " " else key)
        self.repl.paint()  # has some side effects we need to be wary of


class TestWidthAwareness(HigherLevelCurtsiesPaintingTest):
    def test_cursor_position_with_fullwidth_char(self):
        self.repl.add_normal_character("間")

        cursor_pos = self.repl.paint()[1]
        self.assertEqual(cursor_pos, (0, 6))

    def test_cursor_position_with_padding_char(self):
        # odd numbered so fullwidth chars don't wrap evenly
        self.repl.width = 11
        [self.repl.add_normal_character(c) for c in "ｗｉｄｔｈ"]

        cursor_pos = self.repl.paint()[1]
        self.assertEqual(cursor_pos, (1, 4))

    def test_display_of_padding_chars(self):
        self.repl.width = 11
        [self.repl.add_normal_character(c) for c in "ｗｉｄｔｈ"]

        self.enter()
        expected = [">>> ｗｉｄ ", "ｔｈ"]  # <--- note the added trailing space
        result = [d.s for d in self.repl.display_lines[0:2]]
        self.assertEqual(result, expected)


class TestCurtsiesRewindRedraw(HigherLevelCurtsiesPaintingTest):
    def test_rewind(self):
        self.repl.current_line = "1 + 1"
        self.enter()
        screen = [">>> 1 + 1", "2", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (2, 4))
        self.repl.undo()
        screen = [">>> "]
        self.assert_paint_ignoring_formatting(screen, (0, 4))

    def test_rewind_contiguity_loss(self):
        self.enter("1 + 1")
        self.enter("2 + 2")
        self.enter("def foo(x):")
        self.repl.current_line = "    return x + 1"
        screen = [
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> def foo(x):",
            "...     return x + 1",
        ]
        self.assert_paint_ignoring_formatting(screen, (5, 8))
        self.repl.scroll_offset = 1
        self.assert_paint_ignoring_formatting(screen[1:], (4, 8))
        self.undo()
        screen = ["2", ">>> 2 + 2", "4", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (3, 4))
        self.undo()
        screen = ["2", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (1, 4))
        self.undo()
        screen = [
            CONTIGUITY_BROKEN_MSG[: self.repl.width],
            ">>> ",
            "",
            "",
            "",
            " ",
        ]  # TODO why is that there? Necessary?
        self.assert_paint_ignoring_formatting(screen, (1, 4))
        screen = [">>> "]
        self.assert_paint_ignoring_formatting(screen, (0, 4))

    def test_inconsistent_history_doesnt_happen_if_onscreen(self):
        self.enter("1 + 1")
        screen = [">>> 1 + 1", "2", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (2, 4))
        self.enter("2 + 2")
        screen = [">>> 1 + 1", "2", ">>> 2 + 2", "4", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (4, 4))
        self.repl.display_lines[0] = self.repl.display_lines[0] * 2
        self.undo()
        screen = [">>> 1 + 1", "2", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (2, 4))

    def test_rewind_inconsistent_history(self):
        self.enter("1 + 1")
        self.enter("2 + 2")
        self.enter("3 + 3")
        screen = [">>> 1 + 1", "2", ">>> 2 + 2", "4", ">>> 3 + 3", "6", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (6, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[2:], (4, 4))
        self.repl.display_lines[0] = self.repl.display_lines[0] * 2
        self.undo()
        screen = [
            INCONSISTENT_HISTORY_MSG[: self.repl.width],
            ">>> 2 + 2",
            "4",
            ">>> ",
            "",
            " ",
        ]
        self.assert_paint_ignoring_formatting(screen, (3, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[1:-2], (2, 4))
        self.assert_paint_ignoring_formatting(screen[1:-2], (2, 4))

    def test_rewind_inconsistent_history_more_lines_same_screen(self):
        self.repl.width = 60
        sys.a = 5
        self.enter("import sys")
        self.enter("for i in range(sys.a):")
        self.enter("    print(sys.a)")
        self.enter("")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> import sys",
            ">>> for i in range(sys.a):",
            "...     print(sys.a)",
            "... ",
            "5",
            "5",
            "5",
            "5",
            "5",
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (13, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[9:], (4, 4))
        sys.a = 6
        self.undo()
        screen = [
            INCONSISTENT_HISTORY_MSG[: self.repl.width],
            "6",
            # everything will jump down a line - that's perfectly
            # reasonable
            ">>> 1 + 1",
            "2",
            ">>> ",
            " ",
        ]
        self.assert_paint_ignoring_formatting(screen, (4, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[1:-1], (3, 4))

    def test_rewind_inconsistent_history_more_lines_lower_screen(self):
        self.repl.width = 60
        sys.a = 5
        self.enter("import sys")
        self.enter("for i in range(sys.a):")
        self.enter("    print(sys.a)")
        self.enter("")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> import sys",
            ">>> for i in range(sys.a):",
            "...     print(sys.a)",
            "... ",
            "5",
            "5",
            "5",
            "5",
            "5",
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (13, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[9:], (4, 4))
        sys.a = 8
        self.undo()
        screen = [
            INCONSISTENT_HISTORY_MSG[: self.repl.width],
            "8",
            "8",
            "8",
            ">>> 1 + 1",
            "2",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen)
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[-5:])

    def test_rewind_inconsistent_history_more_lines_raise_screen(self):
        self.repl.width = 60
        sys.a = 5
        self.enter("import sys")
        self.enter("for i in range(sys.a):")
        self.enter("    print(sys.a)")
        self.enter("")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> import sys",
            ">>> for i in range(sys.a):",
            "...     print(sys.a)",
            "... ",
            "5",
            "5",
            "5",
            "5",
            "5",
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (13, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[9:], (4, 4))
        sys.a = 1
        self.undo()
        screen = [
            INCONSISTENT_HISTORY_MSG[: self.repl.width],
            "1",
            ">>> 1 + 1",
            "2",
            ">>> ",
            " ",
        ]
        self.assert_paint_ignoring_formatting(screen)
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[1:-1])

    def test_rewind_history_not_quite_inconsistent(self):
        self.repl.width = 50
        sys.a = 5
        self.enter("for i in range(__import__('sys').a):")
        self.enter("    print(i)")
        self.enter("")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> for i in range(__import__('sys').a):",
            "...     print(i)",
            "... ",
            "0",
            "1",
            "2",
            "3",
            "4",
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (12, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[8:], (4, 4))
        sys.a = 6
        self.undo()
        screen = [
            "5",
            # everything will jump down a line - that's perfectly
            # reasonable
            ">>> 1 + 1",
            "2",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (3, 4))

    def test_rewind_barely_consistent(self):
        self.enter("1 + 1")
        self.enter("2 + 2")
        self.enter("3 + 3")
        screen = [">>> 1 + 1", "2", ">>> 2 + 2", "4", ">>> 3 + 3", "6", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (6, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[2:], (4, 4))
        self.repl.display_lines[2] = self.repl.display_lines[2] * 2
        self.undo()
        screen = [">>> 2 + 2", "4", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (2, 4))

    def test_clear_screen(self):
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [">>> 1 + 1", "2", ">>> 2 + 2", "4", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (4, 4))
        self.repl.request_paint_to_clear_screen = True
        screen = [">>> 1 + 1", "2", ">>> 2 + 2", "4", ">>> ", "", "", "", ""]
        self.assert_paint_ignoring_formatting(screen, (4, 4))

    def test_scroll_down_while_banner_visible(self):
        self.repl.status_bar.message("STATUS_BAR")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
            "STATUS_BAR                      ",
        ]
        self.assert_paint_ignoring_formatting(screen, (4, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[1:], (3, 4))

    def test_clear_screen_while_banner_visible(self):
        self.repl.status_bar.message("STATUS_BAR")
        self.enter("1 + 1")
        self.enter("2 + 2")
        screen = [
            ">>> 1 + 1",
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
            "STATUS_BAR                      ",
        ]
        self.assert_paint_ignoring_formatting(screen, (4, 4))
        self.repl.scroll_offset += len(screen) - self.repl.height
        self.assert_paint_ignoring_formatting(screen[1:], (3, 4))

        self.repl.request_paint_to_clear_screen = True
        screen = [
            "2",
            ">>> 2 + 2",
            "4",
            ">>> ",
            "",
            "",
            "",
            "STATUS_BAR                      ",
        ]
        self.assert_paint_ignoring_formatting(screen, (3, 4))

    def test_cursor_stays_at_bottom_of_screen(self):
        """infobox showing up during intermediate render was causing this to
        fail, #371"""
        self.repl.width = 50
        self.repl.current_line = "__import__('random').__name__"
        with output_to_repl(self.repl):
            self.repl.on_enter(new_code=False)
        screen = [">>> __import__('random').__name__", "'random'"]
        self.assert_paint_ignoring_formatting(screen)

        with output_to_repl(self.repl):
            self.repl.process_event(self.refresh_requests.pop())
        screen = [">>> __import__('random').__name__", "'random'", ""]
        self.assert_paint_ignoring_formatting(screen)

        with output_to_repl(self.repl):
            self.repl.process_event(self.refresh_requests.pop())
        screen = [">>> __import__('random').__name__", "'random'", ">>> "]
        self.assert_paint_ignoring_formatting(screen, (2, 4))

    def test_unhighlight_paren_bugs(self):
        """two previous bugs, parent didn't highlight until next render
        and paren didn't unhighlight until enter"""
        self.assertEqual(self.repl.rl_history.entries, [""])
        self.enter("(")
        self.assertEqual(self.repl.rl_history.entries, [""])
        screen = [">>> (", "... "]
        self.assertEqual(self.repl.rl_history.entries, [""])
        self.assert_paint_ignoring_formatting(screen)
        self.assertEqual(self.repl.rl_history.entries, [""])

        with output_to_repl(self.repl):
            self.assertEqual(self.repl.rl_history.entries, [""])
            self.repl.process_event(")")
            self.assertEqual(self.repl.rl_history.entries, [""])
        screen = fsarray(
            [
                cyan(">>> ") + on_magenta(bold(red("("))),
                green("... ") + on_magenta(bold(red(")"))),
            ]
        )
        self.assert_paint(screen, (1, 5))

        with output_to_repl(self.repl):
            self.repl.process_event(" ")
        screen = fsarray(
            [
                cyan(">>> ") + yellow("("),
                green("... ") + yellow(")") + bold(cyan(" ")),
            ]
        )
        self.assert_paint(screen, (1, 6))

    def test_472(self):
        [self.send_key(c) for c in "(1, 2, 3)"]
        with output_to_repl(self.repl):
            self.send_key("\n")
            self.send_refreshes()
            self.send_key("<UP>")
            self.repl.paint()
            [self.send_key("<LEFT>") for _ in range(4)]
            self.send_key("<BACKSPACE>")
            self.send_key("4")
            self.repl.on_enter()
            self.send_refreshes()
        screen = [
            ">>> (1, 2, 3)",
            "(1, 2, 3)",
            ">>> (1, 4, 3)",
            "(1, 4, 3)",
            ">>> ",
        ]
        self.assert_paint_ignoring_formatting(screen, (4, 4))


def completion_target(num_names, chars_in_first_name=1):
    class Class(object):
        pass

    if chars_in_first_name < 1:
        raise ValueError("need at least one char in each name")
    elif chars_in_first_name == 1 and num_names > len(string.ascii_letters):
        raise ValueError("need more chars to make so many names")

    names = gen_names()
    if num_names > 0:
        setattr(Class, "a" * chars_in_first_name, 1)
        next(names)  # use the above instead of first name
    for _, name in zip(range(num_names - 1), names):
        setattr(Class, name, 0)

    return Class()


def gen_names():
    for letters in itertools.chain(
        itertools.combinations_with_replacement(string.ascii_letters, 1),
        itertools.combinations_with_replacement(string.ascii_letters, 2),
    ):
        yield "".join(letters)


class TestCompletionHelpers(TestCase):
    def test_gen_names(self):
        self.assertEqual(
            list(zip([1, 2, 3], gen_names())), [(1, "a"), (2, "b"), (3, "c")]
        )

    def test_completion_target(self):
        target = completion_target(14)
        self.assertEqual(
            len([x for x in dir(target) if not x.startswith("_")]), 14
        )


class TestCurtsiesInfoboxPaint(HigherLevelCurtsiesPaintingTest):
    def test_simple(self):
        self.repl.width, self.repl.height = (20, 30)
        self.locals["abc"] = completion_target(3, 50)
        self.repl.current_line = "abc"
        self.repl.cursor_offset = 3
        self.repl.process_event(".")
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "+------------------+",
        ]
        self.assert_paint_ignoring_formatting(screen, (0, 8))

    def test_fill_screen(self):
        self.repl.width, self.repl.height = (20, 15)
        self.locals["abc"] = completion_target(20, 100)
        self.repl.current_line = "abc"
        self.repl.cursor_offset = 3
        self.repl.process_event(".")
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "| d                |",
            "| e                |",
            "| f                |",
            "| g                |",
            "| h                |",
            "| i                |",
            "| j                |",
            "| k                |",
            "| l                |",
            "+------------------+",
        ]
        self.assert_paint_ignoring_formatting(screen, (0, 8))

    def test_lower_on_screen(self):
        self.repl.get_top_usable_line = lambda: 10  # halfway down terminal
        self.repl.width, self.repl.height = (20, 15)
        self.locals["abc"] = completion_target(20, 100)
        self.repl.current_line = "abc"
        self.repl.cursor_offset = 3
        self.repl.process_event(".")
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "| d                |",
            "| e                |",
            "| f                |",
            "| g                |",
            "| h                |",
            "| i                |",
            "| j                |",
            "| k                |",
            "| l                |",
            "+------------------+",
        ]
        # behavior before issue #466
        self.assert_paint_ignoring_formatting(
            screen, try_preserve_history_height=0
        )
        self.assert_paint_ignoring_formatting(screen, min_infobox_height=100)
        # behavior after issue #466
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "+------------------+",
        ]
        self.assert_paint_ignoring_formatting(screen)

    def test_at_bottom_of_screen(self):
        self.repl.get_top_usable_line = lambda: 17  # two lines from bottom
        self.repl.width, self.repl.height = (20, 15)
        self.locals["abc"] = completion_target(20, 100)
        self.repl.current_line = "abc"
        self.repl.cursor_offset = 3
        self.repl.process_event(".")
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "| d                |",
            "| e                |",
            "| f                |",
            "| g                |",
            "| h                |",
            "| i                |",
            "| j                |",
            "| k                |",
            "| l                |",
            "+------------------+",
        ]
        # behavior before issue #466
        self.assert_paint_ignoring_formatting(
            screen, try_preserve_history_height=0
        )
        self.assert_paint_ignoring_formatting(screen, min_infobox_height=100)
        # behavior after issue #466
        screen = [
            ">>> abc.",
            "+------------------+",
            "| aaaaaaaaaaaaaaaa |",
            "| b                |",
            "| c                |",
            "+------------------+",
        ]
        self.assert_paint_ignoring_formatting(screen)



# encoding: utf-8

from __future__ import unicode_literals

from bpython.test import unittest
from bpython.curtsiesfrontend import parse
from curtsies.fmtfuncs import yellow, cyan, green, bold


class TestExecArgs(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(parse.parse("\x01y\x03print\x04"), yellow("print"))

        self.assertEqual(
            parse.parse(
                "\x01y\x03print\x04\x01c\x03 \x04\x01g\x031\x04\x01c"
                "\x03 \x04\x01Y\x03+\x04\x01c\x03 \x04\x01g\x032\x04"
            ),
            yellow("print")
            + cyan(" ")
            + green("1")
            + cyan(" ")
            + bold(yellow("+"))
            + cyan(" ")
            + green("2"),
        )

    def test_peal_off_string(self):
        self.assertEqual(
            parse.peel_off_string("\x01RI\x03]\x04asdf"),
            (
                {
                    "bg": "I",
                    "string": "]",
                    "fg": "R",
                    "colormarker": "\x01RI",
                    "bold": "",
                },
                "asdf",
            ),
        )



# encoding: utf-8

from __future__ import absolute_import

import gettext
import locale
import os.path
import sys

from .. import package_dir
from .._py3compat import py3

translator = None

if py3:

    def _(message):
        return translator.gettext(message)

    def ngettext(singular, plural, n):
        return translator.ngettext(singular, plural, n)


else:

    def _(message):
        return translator.ugettext(message)

    def ngettext(singular, plural, n):
        return translator.ungettext(singular, plural, n)


def init(locale_dir=None, languages=None):
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error:
        # This means that the user's environment is broken. Let's just continue
        # with the default C locale.
        sys.stderr.write(
            "Error: Your locale settings are not supported by "
            "the system. Using the fallback 'C' locale instead. "
            "Please fix your locale settings.\n"
        )

    global translator
    if locale_dir is None:
        locale_dir = os.path.join(package_dir, "translations")

    translator = gettext.translation(
        "bpython", locale_dir, languages, fallback=True
    )



# encoding: utf-8

import os
from collections import defaultdict

from bpython import importcompletion

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler
except ImportError:

    def ModuleChangedEventHandler(*args):
        return None


else:

    class ModuleChangedEventHandler(FileSystemEventHandler):
        def __init__(self, paths, on_change):
            self.dirs = defaultdict(set)
            self.on_change = on_change
            self.modules_to_add_later = []
            self.observer = Observer()
            self.old_dirs = defaultdict(set)
            self.started = False
            self.activated = False
            for path in paths:
                self._add_module(path)

        def reset(self):
            self.dirs = defaultdict(set)
            del self.modules_to_add_later[:]
            self.old_dirs = defaultdict(set)
            self.observer.unschedule_all()

        def _add_module(self, path):
            """Add a python module to track changes"""
            path = os.path.abspath(path)
            for suff in importcompletion.SUFFIXES:
                if path.endswith(suff):
                    path = path[: -len(suff)]
                    break
            dirname = os.path.dirname(path)
            if dirname not in self.dirs:
                self.observer.schedule(self, dirname, recursive=False)
            self.dirs[dirname].add(path)

        def _add_module_later(self, path):
            self.modules_to_add_later.append(path)

        def track_module(self, path):
            """
            Begins tracking this if activated, or remembers to track later.
            """
            if self.activated:
                self._add_module(path)
            else:
                self._add_module_later(path)

        def activate(self):
            if self.activated:
                raise ValueError("%r is already activated." % (self,))
            if not self.started:
                self.started = True
                self.observer.start()
            for dirname in self.dirs:
                self.observer.schedule(self, dirname, recursive=False)
            for module in self.modules_to_add_later:
                self._add_module(module)
            del self.modules_to_add_later[:]
            self.activated = True

        def deactivate(self):
            if not self.activated:
                raise ValueError("%r is not activated." % (self,))
            self.observer.unschedule_all()
            self.activated = False

        def on_any_event(self, event):
            dirpath = os.path.dirname(event.src_path)
            paths = [path + ".py" for path in self.dirs[dirpath]]
            if event.src_path in paths:
                self.on_change(files_modified=[event.src_path])
